      SUBROUTINE  oto (F, NCOL, NROW, 
     .     par, maxap, lobad, hibad, phpadu, readns,
     .     maxstars, xclist, yclist, VAR, WATCH,
     .     apmaglist, magerrlist, skylist, skysigs, skyskws)
C
C=======================================================================
C
C This subroutine derives the concentric aperture photometry.  At 
C present, this is the only place in all of DAOPHOT where sky values 
C are derived for the individual stars.
C
C               OFFICIAL DAO VERSION:  1996 May 24
C
C Argument
C
C    WATCH (INPUT) governs whether information relating to the progress 
C          of the reductions is to be typed on the terminal screen
C          during execution.
C
C=======================================================================
C
      IMPLICIT NONE
      INTEGER MINSKY, MAXSKY, MAXAP, NCOL, NROW, starno
      integer maxstars
      PARAMETER  (MINSKY=20)
      PARAMETER  (MAXSKY=100000)
c      PARAMETER  (MINSKY=20, MAXAP=12)
C
C Parameters:
C
C MINSKY is the smallest number of pixels from which the sky may be
C        determined.  If for some star the number of sky pixels
C        is less than MINSKY, an error code will result and 
C        control will return to the main program.
C
C MAXSKY the maximum number of pixels allowed in the sky annulus.
C        This and the user's requested inner sky radius will later
C        determine the maximum permitted outer sky radius.
C
C MAXAP  the maximum number of star apertures allowed.
C
c      integer istarlist(maxstars)
      real xclist(maxstars)
      real yclist(maxstars)
      real skylist(maxstars)
      real skysigs(maxstars)
      real skyskws(maxstars)

      CHARACTER*80 LINE
      CHARACTER*30 COOFIL, MAGFIL, PSFFIL, PROFIL, GRPFIL, TABLE
c      CHARACTER SWITCH*30, EXTEND*30, LBL(MAXAP+2)*26, CASE*4
      DOUBLE PRECISION MAGLIM, MAGSQ, WT, SUMWT, DLOG10, DBLE, DMAX1
      DOUBLE PRECISION APMAG(MAXAP), AREA(MAXAP)
      REAL SKY(MAXSKY), MAGERR(MAXAP), BTOT(MAXAP)
c      REAL oraw(NCOL*NROW)
      REAL F(NCOL,NROW), ERROR(3)
      REAL PAR(MAXAP+2), PMIN(MAXAP+2), PMAX(MAXAP+2), RADSQ(MAXAP)
      REAL AVERAG(MAXAP), SIGMAS(MAXAP)
      REAL AMIN1, AMAX1, SQRT, PCTILE
      INTEGER INDEX(MAXSKY)
      INTEGER MIN0, MAX0

      real apmaglist(maxstars,maxap)
      real magerrlist(maxstars,maxap)
C
      REAL A, B, FACTOR
      REAL LOBAD, SKYMOD, SKYSIG, SKYSKW, SIGSQ, SKYVAR
      REAL DATUM, R, RSQ, FRACTN, EDGE, HIBAD, INBAD, THRESH, DUM
      REAL PHPADU, READNS, XC, YC, DMAG, WATCH, APMXSQ, RONOIS
      REAL RINSQ, ROUT, ROUTSQ, DYSQ, XMAX, YMAX
      INTEGER I, J, K, L, NAPER, IDUM, NMAG, LX, LY, NX, NY
      INTEGER ISTAR, MX, MY, NSKY, ISTAT, NL, IANNUL, VAR
C
      COMMON /FILNAM/ COOFIL, MAGFIL, PSFFIL, PROFIL, GRPFIL
C
      DATA FACTOR /1.5/
c      DATA PAR /MAXAP*0., 2*0./, PMIN /1.E-30, MAXAP*0., 1./, 
c     .     PMAX/MAXAP*1.E30, 2*1.E30/
c      DATA TABLE / 'photo.opt                     ' /
c      DATA LBL/' A1  RADIUS OF APERTURE  1',
c     .         ' A2  RADIUS OF APERTURE  2',
c     .         ' A3  RADIUS OF APERTURE  3',
c     .         ' A4  RADIUS OF APERTURE  4',
c     .         ' A5  RADIUS OF APERTURE  5',
c     .         ' A6  RADIUS OF APERTURE  6',
c     .         ' A7  RADIUS OF APERTURE  7',
c     .         ' A8  RADIUS OF APERTURE  8',
c     .         ' A9  RADIUS OF APERTURE  9',
c     .         ' AA  RADIUS OF APERTURE 10',
c     .         ' AB  RADIUS OF APERTURE 11',
c     .         ' AC  RADIUS OF APERTURE 12',
c     .         ' IS       INNER SKY RADIUS',
c     .         ' OS       OUTER SKY RADIUS'/
C
C-----------------------------------------------------------------------
C
C SECTION 1
C
C Ascertain the name of the aperture photometry parameter table, and
C read it in.  Then set up all necessary variables for the forthcoming
C reductions. Finally, identify and open the input and output files.
C
c      L = MAXAP+1
c      PMIN(L) = 1.
c      CALL TBLANK
c      CALL GETNAM ('File with aperture radii:', TABLE)
c      IF (TABLE .EQ. 'END OF FILE') THEN
c         CALL TBLANK
c         RETURN
c      END IF
C
c 1000 CALL OPTION (TABLE, MAXAP+2, LBL, PAR, PMIN, PMAX, 'PHO>', ISTAT)
c      IF (ISTAT .NE. 0) RETURN
C

c        do i=1,ncol
c          do j=1,nrow
c            F(i,j) = oraw((j-1)*ncol+i)
c          end do
c        end do

C Count up the number of apertures that will be used.  The first zero or
C negative number encountered terminates the list.
C
      NAPER=MAXAP
      APMXSQ=-1.
      DO 1010 I=1,MAXAP
      IF (PAR(I) .LE. 0.0) GO TO 1020
      RADSQ(I) = PAR(I)**2
 1010 APMXSQ=AMAX1(APMXSQ, (PAR(I)+0.5)**2)
      GO TO 1030
C
 1020 NAPER=I-1
 1030 CONTINUE
C
C NAPER   is the number of apertures, whose radii are stored in 
C         elements 1 through NAPER of the array PAR.  
C
C APMXSQ  is the outermost edge of the largest aperture-- if the 
C         distance squared of the center of a pixel from the centroid of
C         the star is greater than APMXSQ, then we know that no part 
C         of the pixel is to be included in any aperture.
C
C Now define the other variables whose values are in the table.
C
      RINSQ=AMAX1(PAR(MAXAP+1), 0.)**2      ! Inner sky radius squared 
      ROUTSQ = REAL(MAXSKY)/3.142 + RINSQ
      DUM = PAR(MAXAP+2)**2
c      IF (DUM .GT. ROUTSQ) THEN
c         CALL TBLANK
c         CALL TBLANK
c         CALL STUPID 
c     .        ('   *** You have specified too big a sky annulus. ***')
c         WRITE (6,6) SQRT(ROUTSQ)
c    6    FORMAT (F10.2, ' pixels is the largest outer sky radius ',
c     .        'currently permitted.')
c         RETURN
c      ELSE IF (DUM .LE. RINSQ) THEN
c         CALL TBLANK
c         CALL STUPID 
c     .   ('Your outer sky radius is no bigger than the inner radius.')
c         WRITE (6,8)
c    8    FORMAT ('Please try again.')
c         GO TO 1000
c      ELSE
         ROUT = PAR(MAXAP+2)
         ROUTSQ = DUM
c      END IF
C
C Inquire the name of the input data file with the stellar positions,
C and open it.
C
c      CALL TBLANK
c  900 CALL GETNAM ('File with the positions:', COOFIL)
c      IF ((COOFIL .EQ. 'END OF FILE') .OR. 
c     .     (COOFIL .EQ. 'GIVE UP')) THEN
c         COOFIL = ' '
c         RETURN
c      END IF
c      CALL INFILE (2, COOFIL, ISTAT)
c      IF (ISTAT .NE. 0) THEN
c         CALL STUPID ('Error opening input file '//COOFIL)
c         COOFIL = 'GIVE UP'
c         GO TO 900
c      END IF
c      NL = -1
c      LOBAD = -10.
c      HIBAD = INBAD
c      CALL RDHEAD(2, NL, IDUM, IDUM, LOBAD, HIBAD, THRESH, DUM, 
c     .     PHPADU, READNS, R)
c      IF (NL .LT. 1) NL=1
c      RONOIS = READNS**2
C
C Inquire file name for output aperture photometry results, and open
C the new file.
C
c      MAGFIL=SWITCH(COOFIL, CASE('.ap'))
c  950 CALL GETNAM ('File for the magnitudes:', MAGFIL)
c      IF ((MAGFIL .EQ. 'END OF FILE') .OR. 
c     .     (MAGFIL .EQ. 'GIVE UP')) THEN
c         CALL CLFILE (2)
c         MAGFIL = ' '
c         RETURN
c      END IF
c      MAGFIL = EXTEND(MAGFIL, CASE('ap'))
c      CALL OUTFIL (3, MAGFIL, ISTAT)
c      IF (ISTAT .NE. 0) THEN
c         CALL STUPID ('Error opening output file '//MAGFIL)
c         MAGFIL = 'GIVE UP'
c         GO TO 950
c      END IF
c      CALL WRHEAD (3, 2, NCOL, NROW, 7, LOBAD, HIBAD, THRESH, 
c     .     PAR(1), PHPADU, READNS, R)
C
C If progress is being monitored, type out column headers.
C
      IF (WATCH .GT. 0.5) WRITE (6,610)
  610 FORMAT (/13X, 'STAR', 5X, 'X', 7X, 'Y', 9X, 'MAG.(1)', 8X, 'SKY')
C
C Initialize variables for the computation of the magnitude limit.
C
      MAGLIM=0.0D0
      MAGSQ=0.0D0
      SUMWT=0.0D0
      NMAG=0
C
C-----------------------------------------------------------------------
C
C SECTION 2
C
C Derive aperture photometry object by object.
C
C Get the coordinates of next object to be measured.
C

      starno = 0
      LX = 1
      LY = 1
      NX = NCOL
      NY = NROW
      XMAX = NCOL+0.5
      YMAX = NROW+0.5
cxyz  table = switch(magfil, 'x')                         ! test
cxyz  call coppic (table, istat)                          ! test
c      CALL RDARAY ('DATA', LX, LY, NX, NY, NCOL, F, ISTAT)
c 2000 CALL RDSTAR (2, NL, ISTAR, XC, YC, DMAG, DUM)
 2000 continue
      starno = starno + 1
      IF (starno.eq. maxstars+1) GO TO 9000
c      istar = istarlist(starno)
      istar = starno
      xc = xclist(starno)
      yc = yclist(starno)
      dmag = 0.0
c      dmag = dmaglist(starno)
c      IF (ISTAR .LT. 0) GO TO 9000
      IF (ISTAR .EQ. 0) GO TO 2000
      IF ((XC .LT. 0.5) .OR. (YC .LT. 0.5) .OR. 
     .     (XC .GT. XMAX) .OR. (YC .GT. YMAX)) GO TO 2000
C
C Compute the limits of the submatrix.
C
      LX = MAX0(1, INT(XC-ROUT)+1)
      MX = MIN0(NCOL, INT(XC+ROUT))
      LY = MAX0(1, INT(YC-ROUT)+1)
      MY = MIN0(NROW, INT(YC+ROUT))
      EDGE=AMIN1(XC-0.5, (NCOL+0.5)-XC, YC-0.5, (NROW+0.5)-YC)
C
C EDGE is the distance of the star's centroid from the outermost
C extremum of the array.
C
C Initialize star counts and aperture area.
C
      DO 2010 K=1,NAPER
      APMAG(K) = 0.D0
C 
C If this star aperture extends outside the array, the magnitude
C in this aperture will be no good.
C
      IF (EDGE .LT. PAR(K)) APMAG(K)=-1.1D30            ! Null magnitude
C
C If aperture 2 or greater, determine the average flux value in the
C annulus.
C
      BTOT(K) = 0.
      IF ((K .GT. 1) .AND. (FACTOR .GT. 1.E-20)) THEN
         NSKY = 0
         DO J=LY,MY
            DYSQ = (J-YC)**2
            DO I=LX,MX
               R = SQRT(DYSQ+(I-XC)**2)
               IF ((R .GT. PAR(K-1)-0.35) .AND.
     .             (R .LT. PAR(K)+0.35)) THEN
                  NSKY = NSKY+1
                  SKY(NSKY) = F(I,J)
               END IF
            END DO
         END DO
         IF (NSKY .LE. 0) THEN
            AVERAG(K) = 0.
            SIGMAS(K) = 100.
         ELSE
            A = PCTILE(SKY, NSKY, NINT(0.1537*NSKY+0.5))
            B = PCTILE(SKY, NSKY, NINT(0.8363*NSKY+0.5))
            AVERAG(K) = 0.5*(A+B)
            SIGMAS(K) = AMAX1(0.25*(B-A)**2, RONOIS+AVERAG(K)/PHPADU)*
     .           (2.*FACTOR)**2
         END IF
      END IF
 2010 AREA(K)=0.0D0
C
C Now read through the submatrix, picking out the data we want.
C
      NSKY=0
      skysig = 0.0
      skyvar = 0.0
      sigsq = 0.0
      if (maxsky.eq.0) goto 8100
C
      DO 2130 J=LY,MY
      DYSQ=(J-YC)**2
C
         DO 2125 I=LX,MX
            RSQ=DYSQ+(I-XC)**2
            DATUM=F(I,J)
C
C Is this pixel within the sky annulus?
C
            IF ((RSQ .LT. RINSQ) .OR. (RSQ .GT. ROUTSQ) .OR. 
     .           (NSKY .GT. MAXSKY) .OR. (DATUM .LT. LOBAD) .OR. 
     .           (DATUM .GT. HIBAD)) GO TO 2110
            NSKY=NSKY+1
            SKY(NSKY)=DATUM
C
C The inclusion of partial pixels inside the aperture is done as 
C follows:  if the distance of the center of the current pixel from the
C centroid of the star [radius vector r(i,j)] is exactly equal to the 
C radius of the aperture [R(k)], then one-half of the counts in the
C pixel are included.  If r(i,j) < R(k)-0.5, then the entire pixel is
C included, while if r(i,j) > R(k)+0.5, the pixel is wholly excluded.
C In between, viz. for  R(k)-0.5 < r(i,j) < R(k)+0.5, the fraction of
C the counts included varies linearly.  Therefore a circular aperture
C is approximated by an irregular (not even convex) polygon.
C
C If this pixel falls completely outside the LARGEST aperture, go on
C to the next pixel.  Notice that APMXSQ has actually been defined
C as (R(k)+0.5)**2 for the largest value of R(k), in accordance with
C the formula used for the partial pixels.
C
 2110       CONTINUE
            IF (RSQ .GT. APMXSQ) GO TO 2125
            R=SQRT(RSQ)-0.5
            IANNUL = 1
            IF (NAPER .GT. 1) THEN
               DO K=2,NAPER
                  IF ((RSQ .GT. RADSQ(K-1)) .AND.
     .                (RSQ .LT. RADSQ(K))) THEN
                     IANNUL = K
                     GO TO 2112
                  END IF
               END DO
            END IF
 2112       CONTINUE
            IF ((IANNUL .GT. 1) .AND. (FACTOR .GT. 1.E-20)) THEN
               B = ((DATUM-AVERAG(IANNUL))**2/SIGMAS(IANNUL))**FACTOR
               A = (B*AVERAG(IANNUL)+DATUM)
               B = 1./(B+1.)
               A = A*B
cxyz           lx = i                                                 ! test
cxyz           ly = j                                                 ! test
cxyz           nx = 1                                                 ! test
cxyz           ny = 1                                                 ! test
cxyz           call wraray ('copy', lx, ly, nx, ny, ncol, a, istat)   ! test
            ELSE
               A = DATUM
               B = 1.
            END IF
C
            DO 2120 K=1,NAPER
C
C If this pixel falls completely outside THIS aperture, go on to the
C next aperture.
C
               IF (R .GT. PAR(K)) GO TO 2120
               FRACTN=AMAX1(0.0, AMIN1(1.0,PAR(K)-R))
C
C FRACTN is the fraction of the pixel that falls inside the 
C (irregular) aperture.
C
C If the pixel is bad, set the total counts in this aperture to a number
C so negative that it will never be positive again.
C
               IF ((DATUM .GE. LOBAD) .AND. (DATUM .LE. HIBAD) .AND.
     .              (APMAG(K) .GT. -1.0D20)) THEN
                  APMAG(K) = APMAG(K)+DBLE(FRACTN*A)
                  AREA(K) = AREA(K)+DBLE(FRACTN)
                  IF (IANNUL .EQ. 1) THEN
                     BTOT(K) = BTOT(K)+FRACTN
                  ELSE
                     BTOT(K) = BTOT(K)+FRACTN*B
                  END IF
               ELSE IF (DATUM .GT. HIBAD) THEN
                  APMAG(K) = -1.1D38
               ELSE
                  APMAG(K) = -1.0D30
               END IF
 2120       CONTINUE
 2125    CONTINUE
C
 2130 CONTINUE
C
C We have accumulated the brightnesses of individual sky pixels in the
C one-dimensional array SKY.  Pixels falling above or below the BAD 
C limits have already been eliminated.  Now sort SKY to place the 
C pixels in order of increasing brightness.
C
      IF (NSKY .LT. MINSKY)  THEN
c         CALL STUPID ('There aren''t enough pixels in the sky annulus.')
         if (VAR.eq.0) then
         WRITE (6,*)
     .        ' Are you sure your bad pixel thresholds are all right?'
         WRITE (6,*)
     .        ' If so, then you need a larger outer sky radius.'
          endif
c         CALL TBLANK
c         CALL CLFILE (2)
c         CALL CLFILE (3)
c         RETURN
         goto 8100
      END IF
      CALL QUICK (SKY, NSKY, INDEX)
C
C Obtain the mode, standard deviation, and skewness of the peak in the
C sky histogram.
C
      CALL MMM (SKY, NSKY, HIBAD, READNS, 
     .     DUM, DATUM, SKYMOD, SKYSIG, SKYSKW)
      SKYVAR=SKYSIG**2
      SIGSQ=SKYVAR/FLOAT(NSKY)
      skylist(starno) = skymod
      skysigs(starno) = skysig
      skyskws(starno) = skyskw
C
C SKYMOD has units of (ADU/pixel), and SKYSIG is the pixel-to-pixel
C scatter of SKYMOD, in units of (ADU/pixel).  SKYVAR is the
C variance (square of the standard deviation) of the sky brightness,
C (ADU/pixel)**2, and SIGSQ is the square of the standard error of the 
C mean sky brightness.
C
C Subtract the sky from the integrated brightnesses in the apertures,
C convert the results to magnitudes, and compute standard errors.
C

 8100 continue

      DO 2220 I=1,NAPER
      IF ((APMAG(I) .LT. -1.D31).or.(NSKY .LT. MINSKY)) THEN
         APMAG(I) = 99.999D0
         MAGERR(I) = 9.9999
         GO TO 2220
      END IF
C
C If the modal sky value could not be determined, set the magnitude 
C to 94.999.  
C
      IF (SKYSIG .LT. -0.5) THEN
c         type*,'no sky mode'
          GO TO 2210
      END IF
      if (VAR.eq.0) APMAG(I)=APMAG(I)-DBLE(SKYMOD)*AREA(I)
C
C If the star + sky is fainter than the sky, or if the star aperture 
C extends beyond the limits of the picture, or if there is a bad pixel 
C in the star aperture, set the magnitude to 94.999.
C
      IF (APMAG(I) .LE. 0.0D0) THEN
c        type*,'star fainter than sky'
         GO TO 2210
      END IF
      ERROR(1)=SNGL(AREA(I))*SKYVAR
      ERROR(2)=SNGL(APMAG(I))/PHPADU
      ERROR(3)=SIGSQ*SNGL(AREA(I))**2
C
C These variables ERRORn are the respective variances (squares of the
C mean errors) for: (1) random noise inside the star aperture, including
C readout noise and the degree of contamination by other stars in the
C neighborhood, as estimated by the scatter in the sky values (this
C standard error increases as the square root of the area of the 
C aperture); (2) the Poisson statistics of the observed star brightness;
C (3) the uncertainty of the mean sky brightness (this standard error 
C increases directly with the area of the aperture).
C
      MAGERR(I)=1.0857*SQRT(ERROR(1)+ERROR(2)+ERROR(3))/SNGL(APMAG(I))
      IF (I .GT. 1) MAGERR(I) = MAGERR(I) * DSQRT(AREA(I)/BTOT(I))
      MAGERR(I) = AMIN1(9.9999, MAGERR(I))
      APMAG(I)=25.D0-2.5D0*DLOG10(APMAG(I))
      IF (APMAG(I) .GT. 94.999D0) THEN
c        type*,'star way faint'
         GO TO 2210
      END IF
      apmaglist(starno,i) = apmag(i)
      magerrlist(starno,i) = magerr(i)
      GO TO 2220
 2210 APMAG(I)=94.999D0
      MAGERR(I)=9.9999
      apmaglist(starno,i) = apmag(i)
      magerrlist(starno,i) = magerr(i)
 2220 CONTINUE

C
C Write out the answers.
C      
      IF (WATCH .GT. 0.5) WRITE (6,620)ISTAR, XC, YC, APMAG(1),
     .     MAGERR(1), SKYMOD
  620 FORMAT (11X, I6, 2F8.2, F9.3, ' +-', F6.3, F8.1)
c     WRITE (3,320) ISTAR, XC, YC, (APMAG(I), I=1,NAPER)
c 320 FORMAT (/I6, 14F9.3)
c      WRITE (3,321) SKYMOD, AMIN1(999.99,SKYSIG), 
c     .     AMIN1(999.99, AMAX1(-99.99,SKYSKW)),
c     .     (MAGERR(I), I=1,NAPER)
c  321 FORMAT (4X, F9.3, 2F6.2, F8.4, 11F9.4)
C
      IF (APMAG(1) .GT. 99.D0) GO TO 2000
      WT=(2./(2.-DMAG))*(100./MAGERR(1))**2
      MAGLIM=MAGLIM+WT*(SNGL(APMAG(1))-DMAG)
      MAGSQ=MAGSQ+WT*(SNGL(APMAG(1))-DMAG)**2
      SUMWT=SUMWT+WT
      NMAG=NMAG+1
      GO TO 2000
C
C-----------------------------------------------------------------------
C
C Normal return.  
C
C Estimate magnitude limit, close files, and return.
C
 9000 CONTINUE
c      CALL CLFILE (3)
c      CALL CLFILE (2)
      IF (SUMWT .LE. 0.0) RETURN
      MAGLIM=MAGLIM/SUMWT
      MAGSQ=MAGSQ/SUMWT-MAGLIM**2
      MAGSQ=DSQRT(DMAX1(0.D0, MAGSQ))
      WRITE (LINE,630) MAGLIM, MAGSQ
  630 FORMAT (' Estimated magnitude limit (Aperture 1): ', F4.1,
     .      ' +-', F4.1, ' per star.')
c      CALL STUPID (LINE)
cxyz  call clpic ('copy')                                        ! test
      RETURN
C
C-----------------------------------------------------------------------
C
      END!
